# Configuration file for jupyter-server extensions.
# ------------------------------------------------------------------------------
# Application(SingletonConfigurable) configuration
# ------------------------------------------------------------------------------
# The date format used by logging formatters for %(asctime)s
c.Application.log_datefmt = (  # type:ignore[name-defined]
    "%Y-%m-%d %H:%M:%S Simple_Extensions_Example"
)
